//#define __KERNEL__
//#define MODULE

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/if_ether.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/tcp.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <linux/time.h>
#include <linux/timex.h>
#include <linux/rtc.h>
#include <linux/atomic.h>
#include <linux/mutex.h>

#define FW_NONE	0
#define FW_BLACK	1
#define FW_WHITE	2
#define FW_BLOCK	3

#define OP_APPEND	1
#define OP_CLEAR	2
#define OP_LOCAL 	3

#define WR_NONE	0
#define WR_READ	1
#define WR_WRITE 2

static atomic64_t fw_counter;
DEFINE_MUTEX(fw_filter_mutex);

typedef struct _FilterItem
{
	uint8_t protocol;
	uint32_t dst_addr;
	uint16_t dst_port;
	uint8_t allow;

	uint8_t* domain;
	int domain_len;

	int64_t id;
	struct _FilterItem* next;
} FilterItem;

FilterItem* fw_list_head = NULL;
uint32_t local_addr = 0;

struct nf_hook_ops myhook;

static uint8_t* make_dns_data(const char* domain, int* plen)
{
	int len, i, k = 0, r = 0;
    if (domain == NULL)
	{
		*plen = 0;
		return NULL;
	}
    
	len = strlen(domain);
	if (len == 0)
	{
		*plen = 0;
		return NULL;
	}
    
	uint8_t tmp[256];
	uint8_t* data = kmalloc(len + 10, GFP_KERNEL);
	if (data == NULL)
	{
		*plen = 0;
		return NULL;
	}
		
	for (i = 0; i <= len; i++)
	{
		if (domain[i] == '.' || domain[i] == '\0')
		{
			data[r] = k;
			r++;
			memcpy(&data[r], tmp, k);
			r += k;
			k = 0;
		}
		else
		{
			tmp[k] = domain[i];
			k++;
		}
	}

	*plen = r;
	return data;
}

static int find_in_bytes(const uint8_t* data, int len, const uint8_t* f, int flen)
{
	int i;
	if (len < flen)
		return -1;

	for (i = 0; i < len - flen; i++)
	{
		if (memcmp(&data[i], f, flen) == 0)
			return i;
	}

	return -1;
}

FilterItem* fw_item_new(uint8_t protocol, uint32_t daddr, uint16_t dport, uint8_t allow, char* domain)
{
	FilterItem* item = (FilterItem*)kmalloc(sizeof(FilterItem), GFP_KERNEL);
	if (item == NULL)
	{
		printk("kmalloc failed\n");
		return NULL;
	}

	item->id = atomic64_add_return(1, &fw_counter);
	item->protocol = protocol;
	item->dst_addr = daddr;
	item->dst_port = dport;
	item->allow = allow;
	if (domain != NULL)
	{
		item->domain = make_dns_data(domain, &item->domain_len);
	}
	else
	{
		item->domain_len = 0;
		item->domain = NULL;
	}
	item->next = NULL;

	printk("fw_item_new(protocol: %02x, dst_addr: %08x, dst_port: %02x, allow: %d)\n", protocol, daddr, dport, allow);
	if (domain != NULL)
		printk("    domain: %s\n", domain);
	return item;
}

void fw_item_delete(FilterItem* item)
{
	if (item != NULL)
	{
		if (item->domain != NULL && item->domain_len > 0)
			kfree(item->domain);

		kfree(item);
	}
}

uint32_t fw_item_check(FilterItem* item, struct sk_buff *skb)
{
	struct iphdr *piphdr = ip_hdr(skb);
	struct tcphdr* ptcphdr;
	struct udphdr* pudphdr;
	uint8_t* p1;

	if (piphdr->daddr == 0 || piphdr->daddr == 0x0100007f)
		return NF_ACCEPT;

	if ((piphdr->daddr & 0xffff) == (local_addr & 0xffff))
		return NF_ACCEPT;

	if (item == NULL)
		return UINT_MAX;
	
	if (item->protocol == 0 || piphdr->protocol == item->protocol)
	{
		if (piphdr->protocol == IPPROTO_TCP)
		{
			ptcphdr = tcp_hdr(skb);
			//printk("[mcfw] filter: %x,%x, hdr: %x,%x\n", item->dst_addr, item->dst_port, piphdr->daddr, ptcphdr->dest);
			if ((item->dst_addr == 0 || item->dst_addr == piphdr->daddr) &&
			    (item->dst_port == 0 || item->dst_port == ptcphdr->dest))
			{
				//printk("[mcfw] match TCP\n");
				return item->allow ? NF_ACCEPT : NF_DROP;
			}
		}
		else if (piphdr->protocol == IPPROTO_UDP)
		{
			pudphdr = udp_hdr(skb);
			if ((item->dst_addr == 0 || item->dst_addr == piphdr->daddr) &&
			    (item->dst_port == 0 || item->dst_port == pudphdr->dest))
			{
				//DNS
				if (pudphdr->dest == htons(53) && item->domain != NULL && item->domain_len > 0)
				{
					p1 = skb->data + skb_network_header_len(skb) + 8;
					if (find_in_bytes(p1, pudphdr->len - 8, item->domain, item->domain_len) < 0)
						return UINT_MAX;
				}
				return item->allow ? NF_ACCEPT : NF_DROP;
			}
		}
	}

	return UINT_MAX;
}

void fw_list_append(FilterItem* item, FilterItem* head)
{
	FilterItem* p = head, *p0 = NULL;
	while (p != NULL)
	{
		p0 = p;
		p = p->next;
	}

	if (p0 != NULL)
		p0->next = item;
}

void fw_list_clear(FilterItem** phead)
{
	FilterItem* p = *phead, *p1;
	while (p != NULL)
	{
		p1 = p->next;
		fw_item_delete(p);
		p = p1;
	}
	*phead = NULL;
}

uint32_t fw_list_check(FilterItem* head, struct sk_buff *skb)
{
	uint32_t ret;
	FilterItem* p = head;
	while (p != NULL)
	{
		ret = fw_item_check(p, skb);
		if (ret != UINT_MAX)
			return ret;

		p = p->next;
	}

	return NF_ACCEPT;
}

unsigned int hook_func(void * priv, struct sk_buff *skb, const struct nf_hook_state * state)
{
	uint32_t ret = NF_ACCEPT;
	if (mutex_trylock(&fw_filter_mutex))
	{
		ret = fw_list_check(fw_list_head, skb);
		mutex_unlock(&fw_filter_mutex);
	}
	else
	{
		printk("[mcfw] mutex_trylock failed\n");
	}
	return ret;
}

static ssize_t write_control(struct file * fd, const char __user *buf, size_t len, loff_t *ppos)
{
	uint8_t data[1024];
	FilterItem* item = NULL;

	if (len == 0)
		return 0;
	
	memset(data, 0, 1024);
	
	printk("[mcfw] write_control\n");

	if (copy_from_user(data, buf, len))
	{
		printk("[mcfw] copy_from_user failed\n");
		return 0;
	}

	if (data[0] == OP_APPEND)
	{
		mutex_lock(&fw_filter_mutex);
		if (len >= 9)
		{
			item = fw_item_new(data[1], *(uint32_t*)&data[2], *(uint16_t*)&data[6], data[8], len > 9 ? (char*)&data[9] : NULL);
			if (fw_list_head != NULL)
				fw_list_append(item, fw_list_head);
			else
				fw_list_head = item;

			printk("[mcfw] append filter\n");
		}
		mutex_unlock(&fw_filter_mutex);
	}
	else if (data[0] == OP_CLEAR)
	{
		mutex_lock(&fw_filter_mutex);
		fw_list_clear(&fw_list_head);
		mutex_unlock(&fw_filter_mutex);

		printk("[mcfw] clear filter\n");
	}
	else if (data[0] == OP_LOCAL)
	{
		if (len == 5)
		{
			memcpy(&local_addr, &data[1], 4);
			printk("[mcfw] set localaddr: %x\n", local_addr);
		}
	}
	
	return len;
}

struct file_operations fops = 
{
	.owner = THIS_MODULE, 
	.write = write_control,
};

static dev_t dev;
static struct cdev* fw_dev = NULL;
//static int major;

static int __init initmodule(void)
{
	printk("initialize mc_netfirewall\n");

	atomic64_set(&fw_counter, 0);

	myhook.hook = hook_func;
	myhook.hooknum = NF_INET_POST_ROUTING;
	myhook.pf = PF_INET;
	myhook.priority = NF_IP_PRI_FIRST;
	nf_register_net_hook(&init_net, &myhook);

	if (alloc_chrdev_region(&dev, 0, 1, "mcfw"))
	{
		printk("alloc_chrdev_region failed\n");
		return 0;
	}

	fw_dev = cdev_alloc();
	if (fw_dev == NULL)
	{
		printk("cdev_alloc failed\n");
		unregister_chrdev_region(dev, 1);
		return 0;
	}
	cdev_init(fw_dev, &fops);
	cdev_add(fw_dev, dev, 1);

	// fw_fd = register_chrdev(0, "mcfw", &fops);
	// if (fw_fd < 0) 
	// 	printk("Can't register [mcfw] device file!\n"); 

    return 0;
}

static void __exit cleanupmodule(void)
{
	nf_unregister_net_hook(&init_net, &myhook);
	//unregister_chrdev(fw_fd, "mcfw");

	if (fw_dev)
		cdev_del(fw_dev);

	unregister_chrdev_region(dev, 1);

	mutex_lock(&fw_filter_mutex);
	fw_list_clear(&fw_list_head);
	mutex_unlock(&fw_filter_mutex);
	
	mutex_destroy(&fw_filter_mutex);
    printk("cleanup mc_netfirewall\n");
}

module_init(initmodule);
module_exit(cleanupmodule);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("mineway");

